/*  Copyright (C) MOXA Inc. All rights reserved.

    This software is distributed under the terms of the
    MOXA License.  See the file COPYING-MOXA for details.
*/
/*
    canbus_sja1000.c

    This is for the set of low level dirver functions, particular for sja1000 chip.

    2009-10-15	CF Lin
		new release
*/
#include "os_dependent.h"
#include "canbus_def.h"

#if 0
#define MAX_SECS 1073742

void
do_gettimeofday(struct timeval *p)
{
	static unsigned int carrier=0, pMSec=0;
	unsigned int mSec, seconds;

	mSec = GetTickCount();
	if (mSec < pMSec) carrier++;
	pMSec = mSec;
	seconds = mSec/1000;
	p->tv_sec  = carrier*MAX_SECS + seconds;
	p->tv_msec = mSec%1000;
}
#endif

/*
1. When the <transmit error counter> exceeds the limit of 255, the <bus status bit> is set to 
logic 1 (bus-off), the CAN controller will set the <reset mode bit> to logic 1 (present) and 
an <error warning interrupt> is generated, if enabled. The <transmit error counter> is set to 
127 and the <receive error counter> is cleared. It will stay in this mode until the CPU clears 
the <reset mode bit>. Once this is completed the CAN controller will wait the minimum protocol
defined time (128 occurrences of the bus-free signal) counting down the <transmit error counter>. 
After that the <bus status bit> is cleared (bus-on), the <error status bit> is set to logic 0 (ok), 
the <error counters> are reset and an <error warning interrupt> is generated, if enabled. Reading 
the <TX error counter> during this time gives information about the status of the bus-off 
recovery. (**BUS STATUS**)

2. Errors detected during reception or transmission will effect the error counters according to 
the CAN 2.0B protocol specification. The <error status bit> is set when at least one of the error 
counters has reached or exceeded the CPU warning limit (EWLR). An <error warning interrupt> is 
generated, if enabled. The default value of EWLR after hardware reset is 96. (**ERROR STATUS**)

3. If both the <receive status> and the <transmit status> bits are logic 0 (idle) the CAN-bus  
is idle. If both bits are set the controller is waiting to become idle again. After a hardware 
reset 11 consecutive recessive bits have to be detected until the idle status is reached. After 
bus-off this will take 128 of 11 consecutive recessive bits. (**TXRX STATUS**)

4. The <transmission complete status bit> is set to logic 0 (incomplete) whenever the 
<transmission request bit> or the <self reception request bit> is set to logic 1. The 
<transmission complete status bit> will remain at logic 0 until a message is transmitted 
successfully. (**TCS**)

5. If the CPU tries to write to the transmit buffer when the <transmit buffer status bit> is 
logic 0 (locked), the written byte will not be accepted and will be lost without this being 
indicated. (**TBS**)

6. When a message that is to be received has passed the acceptance filter successfully, the CAN 
controller needs space in the RXFIFO to store the message descriptor and for each data byte which 
has been received. If there is not enough space to store the message, that message is dropped and 
the data overrun condition is indicated to the CPU at the moment this message becomes valid. If this 
message is not completed successfully (e.g. due to an error), no overrun condition is indicated.

7. After reading all messages within the RXFIFO and releasing their memory space with the command 
release receive buffer this bit is cleared.
*/

/*	After an initial download, the contents of the registers should not be changed.
		(1) acceptance code, 
		(2) acceptance mask, 
		(3) bus timing registers 0 and 1, and 
		(4) output control. 
	Therefore these registers may only be accessed when the reset request bit in 
	the control register is set HIGH.

	The reset mode (bit "Reset Request" of the control register) is entered automatically 
	after a hardware reset or when the controller enters the bus-off state (bit "Bus Status"
	of the status register). 

	The operating mode is activated by resetting of the reset request bit in the control register.
*/
/* most of the configuration works are operated when the chip is in the reset mode */
/* receive buffer : 64 bytes */

/* use 4 bits to represent operating_mode/read,operating_mode/write,reset_mode/read,reset_mode/write */
/* mode register */
#define PELI_REGADDR_MODE				0	/* rwrw */
#define PELI_BIT_MOD_SM		(1<<4) /* sleep */
#define PELI_BIT_MOD_AFM	(1<<3) /* acceptance filter */
#define PELI_BIT_MOD_STM	(1<<2) /* self test */
#define PELI_BIT_MOD_LOM	(1<<1) /* listen only */
#define PELI_BIT_MOD_RM			1 /* reset */

/* command register */
#define PELI_REGADDR_COMMAND				1	/* wwww */
#define PELI_BIT_CMR_SRR	(1<<4) /* self reception request */
#define PELI_BIT_CMR_CDO	(1<<3) /* clear data overrun */
#define PELI_BIT_CMR_RRB	(1<<2) /* release receive buffer */
#define PELI_BIT_CMR_AT		(1<<1) /* abort transmission */
#define PELI_BIT_CMR_TR			1 /* transmission request */

/* status register */
#define PELI_REGADDR_STATUS				2	/* r-r- */
#define PELI_BIT_SR_BS		(1<<7) /* bus status */
#define PELI_BIT_SR_ES		(1<<6) /* error */
#define PELI_BIT_SR_TS		(1<<5) /* transmit */
#define PELI_BIT_SR_RS		(1<<4) /* receive */
#define PELI_BIT_SR_TCS		(1<<3) /* transmit complete */
#define PELI_BIT_SR_TBS		(1<<2) /* transmit buffer */
#define PELI_BIT_SR_DOS		(1<<1) /* data overrun */
#define PELI_BIT_SR_RBS			1 /* receive buffer */

/* interrupt register */
#define PELI_REGADDR_INTERRUPT_STATUS	3	/* r-r- */
#define PELI_BIT_IR_BEI		(1<<7) /* bus error */
#define PELI_BIT_IR_ALI		(1<<6) /* arbitration lost */
#define PELI_BIT_IR_EPI		(1<<5) /* error passive */
#define PELI_BIT_IR_WUI		(1<<4) /* wake up */
#define PELI_BIT_IR_DOI		(1<<3) /* data overrun  */
#define PELI_BIT_IR_EI		(1<<2) /* error  */
#define PELI_BIT_IR_TI		(1<<1) /* transmit  */
#define PELI_BIT_IR_RI			1 /* receive  */

/* interrupt enable register */
#define PELI_REGADDR_INTERRUPT_ENABLE	4	/* rwrw */
#define PELI_BIT_IER_BEI	(1<<7) /* bus error */
#define PELI_BIT_IER_ALI	(1<<6) /* arbitration lost */
#define PELI_BIT_IER_EPI	(1<<5) /* error passive */
#define PELI_BIT_IER_WUI	(1<<4) /* wake up */
#define PELI_BIT_IER_DOI	(1<<3) /* data overrun  */
#define PELI_BIT_IER_EI		(1<<2) /* error  */
#define PELI_BIT_IER_TI		(1<<1) /* transmit  */
#define PELI_BIT_IER_RI			1 /* receive  */

/* error interrupts */
#define	PELI_INTERRUPT_ERR					0xFC

#define PELI_REGADDR_RESERVED_1			5	/* r-r- */
#define PELI_REGADDR_BUS_TIMING_0		6	/* r-rw */
#define PELI_REGADDR_BUS_TIMING_1		7	/* r-rw */
#define PELI_REGADDR_OUTPUT_CONTROL		8	/* r-rw */
#define PELI_REGADDR_TEST				9	/* rwrw */
#define PELI_REGADDR_RESERVED_2			10	/* r-r- */
#define PELI_REGADDR_ARBITRATION_LOST	11	/* r-r- */
#define PELI_REGADDR_ERROR_CODE			12	/* r-r- */
#define PELI_REGADDR_WARNIMG_LIMIT		13	/* r-rw */
#define PELI_REGADDR_ERROR_COUNTER_RX	14	/* r-rw */
#define PELI_REGADDR_ERROR_COUNTER_TX	15	/* r-rw */
#define PELI_REGADDR_FRAME_INFORMATION	16	/* rw-- */
#define PELI_REGADDR_ACCEPTANCE_CODE	16	/* --rw */
#define PELI_REGADDR_MESSAGE_IDENTIFER	17	/* rw-- */
#define PELI_REGADDR_MESSAGE_DATA_SFF	19	/* rw-- */
#define PELI_REGADDR_ACCEPTANCE_MASK	20	/* --rw */
#define PELI_REGADDR_MESSAGE_DATA_EFF	21	/* rw-- */
#define PELI_REGADDR_MESSAGE_COUNT_RX	29	/* r-r- */
#define PELI_REGADDR_BUFFER_ADDRESS_RX	30	/* r-r- */
#define PELI_REGADDR_CLOCK_DIVIDER		31	/* rwrw */

typedef struct _BUSTIMING
{
	unsigned int bus_timing;
	unsigned char btr0;
	unsigned char btr1;
} BUSTIMING;

/* the list of bus timing items that are supported */
static BUSTIMING bus_timing_tab[] = 
{
	{ 5,	0xBF, 0xFF },
	{ 10,	0x31, 0x1C },
	{ 20,	0x18, 0x1C },
	{ 40,	0x87, 0xFF },
	{ 50,	0x09, 0x1C },
	{ 80,	0x83, 0xFF },
	{ 100,	0x04, 0x1C },
	{ 125,	0x03, 0x1C },
	{ 200,	0x81, 0xFA },
	{ 250,	0x01, 0x1C },
	{ 400,	0x80, 0xFA },
	{ 500,	0x00, 0x1C },
	{ 666,	0x80, 0xB6 },
	{ 800,	0x00, 0x16 },
	{ 1000, 0x00, 0x14 },
};

static unsigned int bus_timing_tab_size = sizeof(bus_timing_tab)/sizeof(bus_timing_tab[0]);

/*	a common used function of setting bits on a register 
	Inputs:
		<port> the hardware port
		<regadr> register address
		<rval> the value to be set on the register
	Returns:
		None
*/
static void 
sja1000_set_regval(HWPORT *port, unsigned char regadr, unsigned char rval)
{
	unsigned char regtmp;

	reg_addr_read( port, regadr, regtmp); 
	regtmp |= rval;
	reg_addr_write(port, regadr, regtmp );
}

/*	a common used function of clearing bits on a register 
	Inputs:
		<port> the hardware port
		<regadr> register address
		<rval> the value to be cleared on the register
	Returns:
		None
*/
static void 
sja1000_clr_regval(HWPORT *port, unsigned char regadr, unsigned char rval)
{
	unsigned char regtmp;

	reg_addr_read( port, regadr, regtmp); 
	regtmp &= ~rval;
	reg_addr_write(port, regadr, regtmp );
}

/* set/clear bits of the mode register */
#define sja1000_set_mode(p,v)				sja1000_set_regval(p,PELI_REGADDR_MODE,v)
#define sja1000_clr_mode(p,v)				sja1000_clr_regval(p,PELI_REGADDR_MODE,v)

/* set/clear bits of the interrupt register */
#define sja1000_set_interrupts(p,v)			sja1000_set_regval(p,PELI_REGADDR_INTERRUPT_ENABLE,v)
#define sja1000_clr_interrupts(p,v)			sja1000_clr_regval(p,PELI_REGADDR_INTERRUPT_ENABLE,v)

/* configuration == reset mode */
#define sja1000_reset_mode(p)				sja1000_set_mode(p,PELI_BIT_MOD_RM) 
#define sja1000_operating_mode(p)			sja1000_clr_mode(p,PELI_BIT_MOD_RM)

/* set/clear bits of the command register */
#define sja1000_set_command(p,v)			reg_addr_write(p,PELI_REGADDR_COMMAND,v)

/* read the status */
#define sja1000_get_status(p,v)				reg_addr_read(p,PELI_REGADDR_STATUS,v)

void
hwport_read_registers(HWPORT *port, unsigned char *regs, unsigned int size)
{
	unsigned int i;

	for (i=0; i < size; i++) reg_addr_read( port, i, regs[i]);
}	

/*	get the register values for a bus timing
	Inputs:
		<timing> the bus timing to be lookuped
	Outputs:
		<btr0><btr1> register values representing the bus timing 
	Returns:
		0 if one is find, otherwise returns -1
*/
static int
sja1000_lookup_bus_timing_registers(unsigned int timing, unsigned char *btr0, unsigned char *btr1)
{
	unsigned int i;

	for (i=0; i < bus_timing_tab_size; i++)
	{
		if (bus_timing_tab[i].bus_timing != timing)
			continue;
		*btr0 = bus_timing_tab[i].btr0;
		*btr1 = bus_timing_tab[i].btr1;
		return 0;
	}
	return -1;
}

/*	get the bus timing of a port by register values
	Inputs:
		<timing> the bus timing to be lookuped
	Outputs:
		<btr0><btr1> register values representing the bus timing 
	Returns:
		0 if one is find, otherwise returns -1
*/
static unsigned int
sja1000_lookup_bus_timing(unsigned char btr0, unsigned char btr1)
{
	unsigned int i;

	for (i=0; i < bus_timing_tab_size; i++)
	{
		if (bus_timing_tab[i].btr0==btr0 &&
			bus_timing_tab[i].btr1==btr1 )
			return bus_timing_tab[i].bus_timing;
	}
	return 0;
}

/*	get parameters of a harwdware port
	Inputs:
		<port> the port
		<arg> the container of the parameters
	Returns:
		0 on success
*/


int
hwport_get_parameters(HWPORT *port, void *arg)
{
	unsigned char rval, regadr;
	unsigned char btr0, btr1;
	unsigned int i, code, mask;
	CANPRM *param = (CANPRM*) arg;

	if (!param || !port)
		return -1;
	memset(param, 0, sizeof(CANPRM));

	/* should enter reset mode */
	sja1000_reset_mode(port);

	/* get bus timing */
	reg_addr_read(port, PELI_REGADDR_BUS_TIMING_0, btr0);
	reg_addr_read(port, PELI_REGADDR_BUS_TIMING_1, btr1);
	param->bus_timing = sja1000_lookup_bus_timing(btr0, btr1);
	if (param->bus_timing==0) {
		sja1000_operating_mode(port);
		return -2;
	}
	/* single or dual mode */
	reg_addr_read(port, PELI_REGADDR_MODE, rval);
	param->acceptance_filter_mode = (rval & PELI_BIT_MOD_AFM)? 1:0;
	/* get acceptance_filter_code, 4 bytes */
	code = 0;
	for (i=0; i < 4; i++)
	{
		/* code */
		regadr = PELI_REGADDR_ACCEPTANCE_CODE+i;
		reg_addr_read(port, regadr, rval);
		code = (code<<8) + rval;
	}
	param->acceptance_filter_code = code;
	/* get acceptance_filter_mask, 4 bytes */
	mask = 0;
	for (i=0; i < 4; i++)
	{
		/* mask */
		regadr = PELI_REGADDR_ACCEPTANCE_MASK+i;
		reg_addr_read(port, regadr, rval);
		mask = (mask<<8) + rval;
	}
	param->acceptance_filter_mask = mask;

	/* return to operating mode */
	sja1000_operating_mode(port);

	return 0;
}

/*	set parameters onto a harwdware port in the reset mode
	Inputs:
		<port> the port
		<arg> the container of the parameters
	Returns:
		0 on success
*/
int
hwport_set_parameters(HWPORT *port, void *arg)
{
	CANPRM *param = (CANPRM*) arg;
	unsigned char btr0, btr1;
	unsigned char rval, regadr;
	unsigned int code, mask;
	int i;

	if (!param || !port)
		return -1;
	if (sja1000_lookup_bus_timing_registers(param->bus_timing, &btr0, &btr1))
		return -2;
	/* setting bus timing must be under the reset mode */
	sja1000_reset_mode(port);
	reg_addr_write(port, PELI_REGADDR_BUS_TIMING_0, btr0 );
	reg_addr_write(port, PELI_REGADDR_BUS_TIMING_1, btr1 );
	if (param->acceptance_filter_mode) /* single_mode */
	{
		sja1000_set_mode(port, PELI_BIT_MOD_AFM);
		/* SFF: ID28~ID21, ID20~ID18[RTR]XXXX, DB1.7~DB1.0, DB2.7~DB2.0 */
		/* EFF: ID28~ID21, ID20~ID13, ID12~ID05, ID04~ID00XXX */
	}
	else
	{
		/* filter 1 : [ACR0,ACR1]&[AMR0,AMR1] and filter 2 : [ACR2,ACR3]&[AMR2,AMR3] */
		sja1000_clr_mode(port, PELI_BIT_MOD_AFM);
		/* SFF: ID28~ID21, ID20~ID18[RTR]DB1.7~DB1.4, DB1.3~DB1.0 */
		/* EFF: ID28~ID21, ID20~ID13 */
	}
	code = param->acceptance_filter_code;
	/* set code/mask, each has 4 bytes */
	for (i=3; i >= 0; i--)
	{
		/* code */
		regadr = PELI_REGADDR_ACCEPTANCE_CODE+i;
		rval = (code & 0xff);
		reg_addr_write(port, regadr, rval);
		code >>= 8;
	}
	mask = param->acceptance_filter_mask;
	/* set code/mask, each has 4 bytes */
	for (i=3; i >= 0; i--)
	{
		/* mask */
		regadr = PELI_REGADDR_ACCEPTANCE_MASK+i;
		rval = (mask & 0xff);
		reg_addr_write(port, regadr, rval);
		mask >>= 8;
	}
	sja1000_operating_mode(port);
	return 0;
}

/*	purge rxbuf of a hardware port
	Inputs:
		<port> the port
	Returns:
		None 
*/
void
hwport_purge_rxbuf(HWPORT *port)
{
	unsigned char count;
	unsigned int i = 0;

	while(i++ < 10) /* few packets most */
	{
		reg_addr_read( port, PELI_REGADDR_MESSAGE_COUNT_RX, count);
		if (count==0)
			break;
		sja1000_set_command(port, PELI_BIT_CMR_RRB);
	}
}

/*	purge txbuf of a hardware port
	Inputs:
		<port> the port
	Returns:
		None 
*/
void
hwport_purge_txbuf(HWPORT *port)
{
	unsigned int i = 0;

	sja1000_set_command(port, PELI_BIT_CMR_AT);
	while(i++ < 10) /* few packets most */
	{
		/* make sure a message is aborted successfully */
		sja1000_get_status(port,port->sr_reg); 
		if (port->sr_reg & PELI_BIT_SR_TCS)
			break;
	}
}

/*	convert interrupts of a port to user types
	Inputs:
		<port> the port
	Returns:
		the bit-wise combination of interrupt types
	Note: 
		refer to 6.4.6. the error bits are 0xfc ..
*/
unsigned int 
hwport_get_interrupts(HWPORT *port)
{
	unsigned char ir_reg = 0;

	/* interrupt register */
	reg_addr_read( port, PELI_REGADDR_INTERRUPT_STATUS, port->ir_reg);
	if( port->ir_reg & PELI_INTERRUPT_ERR) 	ir_reg |= INTERRUPT_TYPE_ERR;
	if( port->ir_reg & PELI_BIT_IR_RI ) 	ir_reg |= INTERRUPT_TYPE_RX;
	if( port->ir_reg & PELI_BIT_IR_TI ) 	ir_reg |= INTERRUPT_TYPE_TX;

	return ir_reg;
}

/*	enable interrupts of a port by complying user types
	Inputs:
		<port> the port
		<ier> the bit-wise combination of interrupt types
	Returns:
		None
*/
static void
hwport_enable_interrupts(HWPORT *port, unsigned int type)
{
	unsigned int rval = 0;

	if( type & INTERRUPT_TYPE_RX)	rval |= PELI_BIT_IER_RI;
	if( type & INTERRUPT_TYPE_TX)	rval |= PELI_BIT_IER_TI;
	if( type & INTERRUPT_TYPE_ERR)	rval |= PELI_INTERRUPT_ERR;
	sja1000_set_interrupts( port, rval );
}

/*	clear interrupts by complying user types of a port
	Inputs:
		<port> the port
		<ier> the bit-wise combination of interrupt types
	Returns:
		None
*/
void
hwport_clear_interrupts(HWPORT *port, unsigned int type)
{
	unsigned int rval = 0;

	if( type & INTERRUPT_TYPE_RX)	rval |= PELI_BIT_IER_RI;
	if( type & INTERRUPT_TYPE_TX)	rval |= PELI_BIT_IER_TI;
	if( type & INTERRUPT_TYPE_ERR)	rval |= PELI_INTERRUPT_ERR;
	sja1000_clr_interrupts( port, rval );
}


/*	start a hardware port
	Inputs:
		<port> the port
	Returns:
*/
static void 
hwport_start( HWPORT *port )
{
	int i = 0;

	/* read out existing interrutp */
	hwport_get_interrupts(port);
	/* go to operating mode */
	sja1000_operating_mode(port);
	/*	at the time of reset, pointers of RXFIFO are reset to initial values. 
		A subsequent read would get parts of the old data-undefined data. 
		So, clear the buffer.
	*/
	hwport_purge_txbuf(port);
	hwport_purge_rxbuf(port);
	mdelay(100); /* wait a bit */
	while(i++ < 100)
	{
		sja1000_get_status(port,port->sr_reg);
		if( (port->sr_reg & (PELI_BIT_SR_BS|PELI_BIT_SR_ES))==0 ) 
			break;
	}
	hwport_enable_interrupts(port,INTERRUPT_TYPE_RX|INTERRUPT_TYPE_TX|INTERRUPT_TYPE_ERR);
}

/*	release a CAN port
	Inputs:
		<port> the port
	Returns:
		None 
*/
void 
hwport_release( HWPORT *port )
{
	sja1000_reset_mode(port);
	hwport_clear_interrupts(port, INTERRUPT_TYPE_RX|INTERRUPT_TYPE_TX|INTERRUPT_TYPE_ERR);
}

/*	initialize a hardware port
	Inputs:
		<port> the port
	Returns:
		0 on success
*/
int 
hwport_init( HWPORT *port )
{
	int ret;
	CANPRM param;
	
	/* read/write in reset mode */
	hwport_release(port);
	/* OCTP0 ((1<<4)), OCTN0(0x80), normal output (0x02) */
	reg_addr_write(port, PELI_REGADDR_OUTPUT_CONTROL, 0xFA); 
	/* setting clock divider register : select PeliCAN mode (0x80) and clock-off (0x08) = 0xCF*/
	reg_addr_write(port, PELI_REGADDR_CLOCK_DIVIDER, (0x88|0x47));
	/* parameters of the CAN bus */
	param.bus_timing = 1000; /* 1000KHz */
	param.acceptance_filter_mode = 0; /* non-zero is for single mode, default is dual mode */
	param.acceptance_filter_code = 0; /* default is 0 */
	param.acceptance_filter_mask = 0xFFFFFFFF; /* default is 0xFFFFFFFF */
	ret = hwport_set_parameters(port, &param);
	if (ret==0)
	{
		hwport_start(port);
	}
	return ret;
}

int
hwport_read_count(HWPORT *port)
{
	unsigned char count;
	
	reg_addr_read( port, PELI_REGADDR_MESSAGE_COUNT_RX, count);
	return (int) count;
}

/*	write data onto the hardware port
	Inputs:
		<port> the port
		<data> data 
		<size> the size of the data
	Returns:
		a packet of a data message
*/
int
hwport_write(HWPORT *port, char *data, int size)
{
	CANMSG *msg;
	unsigned char rval;
	unsigned int offs, i;

	msg = (CANMSG*) data;
	if (size < sizeof(CANMSG)) {
		return 0;
	}

#if 0
	/* according to spec. make sure a message is transmitted successfully */
	do
	{
		/* check status register */
		sja1000_get_status(port,port->sr_reg);
	} while((port->sr_reg & PELI_BIT_SR_TCS)==0 && i++ < 10);
#endif

	/* check status register */
	sja1000_get_status(port,port->sr_reg);

	if((port->sr_reg & PELI_BIT_SR_TBS)==0) 
	{
		//p("error: data in transmitting or bus error");
		return 0; /* data in transmitting or bus error */
	}
	if (msg->length > MAX_DATA_SIZE) msg->length = MAX_DATA_SIZE;
	rval = ((msg->flags & EXTENDED_FRAME)? 0x80:0)|((msg->flags & REMOTE_FRAME)? 0x40:0)|(msg->length);
	reg_addr_write( port, PELI_REGADDR_FRAME_INFORMATION, rval );
	if ( msg->flags & EXTENDED_FRAME) /* EFF */
	{
		rval = (msg->identifier>>21);
		reg_addr_write(port, PELI_REGADDR_MESSAGE_IDENTIFER, rval);
		rval = ((msg->identifier>>13) & 0xFF);
		reg_addr_write(port, (PELI_REGADDR_MESSAGE_IDENTIFER+1), rval);  
		rval = ((msg->identifier>>5) & 0xFF);
		reg_addr_write(port, (PELI_REGADDR_MESSAGE_IDENTIFER+2), rval);
		rval = ((msg->identifier<<3) & 0xFF);
		reg_addr_write(port, (PELI_REGADDR_MESSAGE_IDENTIFER+3), rval);
		offs = PELI_REGADDR_MESSAGE_DATA_EFF;
	}
	else /* SFF */
	{
		rval = ((msg->identifier>>3) & 0xFF);
		reg_addr_write(port, PELI_REGADDR_MESSAGE_IDENTIFER, rval);
		rval = ((msg->identifier<<5) & 0xFF);
		reg_addr_write(port, (PELI_REGADDR_MESSAGE_IDENTIFER+1), rval);  
		offs = PELI_REGADDR_MESSAGE_DATA_SFF;
	}
	
	for( i=0; i < msg->length; i++, offs++ ) 
	{
		reg_addr_write(port, offs, msg->data[i]);
	}
	sja1000_set_command(port, PELI_BIT_CMR_TR ); /* should it be set? transmission request */
	return sizeof(CANMSG);
}

/*	read data from the hardware port
	Inputs:
		<port> the port
		<buf> buffer for storing data 
		<max> the length of the buffer
	Returns:
		the number of bytes to be read
*/
int
hwport_read(HWPORT *port, char *buf, int max)
{
	register unsigned char rval;
	register CANMSG *msg = (CANMSG*) buf;
	unsigned int offs, i;
	int bytes = 0;
	unsigned char count=0;
	CANBST	*stat = (CANBST*) port->private_data;
	
	if (max < sizeof(CANMSG))
		return 0;
	do
	{
		/* check status register */
		sja1000_get_status(port,port->sr_reg);
		if( (port->sr_reg & PELI_BIT_SR_RBS)==0 ) //			(port->sr_reg & (PELI_BIT_SR_TS|PELI_BIT_SR_BS)) )
			break;
		/* check the frame format and data length */
		reg_addr_read( port, PELI_REGADDR_FRAME_INFORMATION, rval );
		msg->length = rval & 0xF; /* length in 4 LSB bits */
		if (msg->length > MAX_DATA_SIZE) 
			msg->length = MAX_DATA_SIZE;
		msg->identifier = 0;
		msg->flags = (rval&0x40)? REMOTE_FRAME:0;
		if ( rval & 0x80 ) /* MSB bit:extended frame */
		{
			msg->flags |= EXTENDED_FRAME; 
			/* identifier, ID28~ID21, ID20~ID13, ID12~ID5, ID4~ID0 */
			reg_addr_read( port, PELI_REGADDR_MESSAGE_IDENTIFER, rval); 
			msg->identifier  = (rval << 21);
			reg_addr_read( port, (PELI_REGADDR_MESSAGE_IDENTIFER+1), rval); 
			msg->identifier |= (rval << 13);
			reg_addr_read( port, (PELI_REGADDR_MESSAGE_IDENTIFER+2), rval); 
			msg->identifier |= (rval << 5);
			reg_addr_read( port, (PELI_REGADDR_MESSAGE_IDENTIFER+3), rval); 
			msg->identifier |= (rval >> 3);
			offs = PELI_REGADDR_MESSAGE_DATA_EFF;
		} 
		else /* standard frame */
		{
			reg_addr_read( port, PELI_REGADDR_MESSAGE_IDENTIFER, rval); 
			msg->identifier  = (rval << 3);
			reg_addr_read( port, (PELI_REGADDR_MESSAGE_IDENTIFER+1), rval); 
			msg->identifier |= (rval >> 5);
			offs = PELI_REGADDR_MESSAGE_DATA_SFF;
		}
		/* get data of the message*/
		for ( i=0; i < msg->length; i++, offs++ ) 
		{
			reg_addr_read( port, offs, msg->data[i] );
		}
//		do_gettimeofday(&msg->timestamp);
		/*	clear receive buffer, 
			1. decrease RX counter
			2. clear RI, otherwise, the RI is still there 
		*/
		sja1000_set_command(port, PELI_BIT_CMR_RRB);
		/* important to read the status here */
		sja1000_get_status(port,port->sr_reg);


		if (port->sr_reg & PELI_BIT_SR_DOS)
		{
			sja1000_set_command(port, PELI_BIT_CMR_CDO);
			if (stat) stat->data_overrun_count++;
		}
		/* move pointers */
		msg++;
		bytes	+= sizeof(CANMSG);
		max		-= sizeof(CANMSG);
	} while(max >= sizeof(CANMSG) && count++ < 10); /* wade: read until max < CANMSG .. */
	return bytes;
}

/*	handle errors of a port
	Inputs:
		<port> the port
	Returns:
		none
*/
void
hwport_handle_errors(HWPORT *port)
{
	unsigned char rval;
	CANBST	*stat = (CANBST*) port->private_data;
/*
1. When the <transmit error counter> exceeds the limit of 255, the <bus status bit> is set to 
logic 1 (bus-off), the CAN controller will set the <reset mode bit> to logic 1 (present) and 
an <error warning interrupt> is generated, if enabled. The <transmit error counter> is set to 
127 and the <receive error counter> is cleared. It will stay in this mode until the CPU clears 
the <reset mode bit>. Once this is completed the CAN controller will wait the minimum protocol
defined time (128 occurrences of the bus-free signal) counting down the <transmit error counter>. 
After that the <bus status bit> is cleared (bus-on), the <error status bit> is set to logic 0 (ok), 
the <error counters> are reset and an <error warning interrupt> is generated, if enabled. Reading 
the <TX error counter> during this time gives information about the status of the bus-off 
recovery. (**BUS STATUS**)
*/

	/* Note: see status register */
	sja1000_get_status(port, port->sr_reg);
	p("get status = %x", port->sr_reg);
	if(port->sr_reg & PELI_BIT_SR_BS ) /* could be <error warning interrupt> */
	{
	}
	if(port->sr_reg & PELI_BIT_SR_ES) 
	{
	}	
	if( port->ir_reg & PELI_BIT_IR_DOI ) /* data overrun of the RXFIFO, the message comes valid again */
	{
		if (stat) stat->data_overrun_count++;
		sja1000_set_command(port, PELI_BIT_CMR_CDO);
	}
	if( port->ir_reg & PELI_BIT_IR_EI ) /* error warning interrupt */
	{
		if(port->sr_reg & PELI_BIT_SR_BS || port->sr_reg & PELI_BIT_SR_ES)
		{
			if (stat) stat->warning_error_count++;
		}
	}
	if( port->ir_reg & PELI_BIT_IER_BEI ) /* bus status */
	{
		if (stat) stat->bus_error_count++;
	}
	if( port->ir_reg & PELI_BIT_IER_ALI ) /* change of error/bus status */
	{
		if (stat) stat->status_error_count++;
	}
	/* restart the port if if is on reset mode */
	reg_addr_read( port, PELI_REGADDR_MODE, rval); 
	if (rval & PELI_BIT_MOD_RM)
	{
		hwport_start(port); /* clear reset mode */
	}
	else
	{
		/* PELI_BIT_SR_ES and PELI_BIT_SR_BS should be 0; */
	}
}

/*	other ioctl for a hardware port
	Inputs:
		<port> the port to access ioctl
		<code> command code
		<ibuf> the input buffer
		<ilen> the length of the input buffer in WIN32, 0 in Linux
	Outputs:
		<obuf> the output buffer (in Linux, it is same as ibuf)
		<olen> the length of the output buffer in WIN32, 0 in Linux
		<rlen> the # of bytes to be returned (NULL in Linux)
	Returns:
		0 on success
*/
int	hwport_ioctl(HWPORT *port, unsigned int code, 
				 void *ibuf, unsigned int ilen, 
				 void *obuf, unsigned int olen, 
				 unsigned int *rlen)
{
	int		ret=0;
	char	param[1024];
	unsigned int size=0;
	CANBST	*stat;

	switch ( code )
	{
	case IOCTL_SET_PARAMETERS:	/* set chip parameters */
		size = sizeof(CANPRM);
		if (!access_ok(VERIFY_READ, ibuf, size))
			ret = -1;
		else
		{
			copy_from_user( param, ibuf, size);
			if (hwport_set_parameters(port, param))
			{
				ret = -2;
			}
		}
		break;
	case IOCTL_GET_PARAMETERS:	/* get current chip parameters */
		size = sizeof(CANPRM);
		if (!access_ok(VERIFY_WRITE, obuf, size))
			ret = -1;
		else if (hwport_get_parameters(port, param))
			ret = -2;
		else
			copy_to_user( obuf, param, size);
		break;
	case IOCTL_GET_DEVICE_STATISTIC:
		size = sizeof(CANBST);
		stat = (CANBST*) port->private_data;
		if (!access_ok(VERIFY_WRITE, obuf, size)){
			ret = -1;
		} else {
			copy_to_user( obuf, stat, size);
		}

		break;
	case IOCTL_GET_DEVICE_REGISTERS:
		size = 32;
		if (!access_ok(VERIFY_WRITE, obuf, size))
			ret = -1;
		else
		{
			hwport_read_registers(port, (unsigned char*) param, size);
			copy_to_user( obuf, param, size);
		}
		break;
	default:
		ret = -9;
		break;
	}
	if (rlen) *rlen = size;
	return ret;
}
